from pathlib import Path
from unittest import TestCase

from rtctools_heat_network.util import run_heat_network_optimization


class TestWarmingUpUnitCases(TestCase):
    def test_1a(self):
        import models.unit_cases.case_1a.src.run_1a as run_1a
        from models.unit_cases.case_1a.src.run_1a import HeatProblem, QTHProblem

        base_folder = Path(run_1a.__file__).resolve().parent.parent

        # Just a "problem is not infeasible"
        _heat_problem, _qth_problem = run_heat_network_optimization(
            HeatProblem, QTHProblem, base_folder=base_folder
        )

    def test_2a(self):
        import models.unit_cases.case_2a.src.run_2a as run_2a
        from models.unit_cases.case_2a.src.run_2a import HeatProblem, QTHProblem

        base_folder = Path(run_2a.__file__).resolve().parent.parent

        # Just a "problem is not infeasible"
        _heat_problem, _qth_problem = run_heat_network_optimization(
            HeatProblem, QTHProblem, base_folder=base_folder
        )

    def test_3a(self):
        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblem, QTHProblem

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        # Just a "problem is not infeasible"
        _heat_problem, _qth_problem = run_heat_network_optimization(
            HeatProblem, QTHProblem, base_folder=base_folder
        )



from rtctools_heat_network.pycml import Connector, Variable

from ._internal import HeatComponent


class HeatPort(HeatComponent, Connector):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(Variable, "Heat")
        self.add_variable(Variable, "Q")
        self.add_variable(Variable, "H")



from numpy import nan

from rtctools_heat_network.pycml import Variable

from ._non_storage_component import _NonStorageComponent


class Source(_NonStorageComponent):
    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(
                    Heat_in=dict(min=0.0, max=0.0),
                    Heat_out=dict(min=0.0),
                ),
                modifiers,
            ),
        )

        self.component_type = "source"

        self.price = nan  # TODO: delete not needed anymore
        self.co2_coeff = 1.0

        # Assumption: heat in/out and added is nonnegative
        # Heat in the return (i.e. cold) line is zero
        self.add_variable(Variable, "Heat_source", min=0.0, nominal=self.Heat_nominal)
        self.add_variable(Variable, "Emission", min=0.0, nominal=self.Heat_nominal)
        self.add_variable(Variable, "dH", min=0.0)

        self.add_equation(self.dH - (self.HeatOut.H - self.HeatIn.H))

        self.add_equation(
            (self.HeatOut.Heat - (self.HeatIn.Heat + self.Heat_source)) / self.Heat_nominal
        )
        # self.add_equation(
        #     (self.Emission - self.Heat_source * self.co2_coeff)
        #     / (self.Heat_nominal * self.co2_coeff)
        # )

        self.add_equation((self.Heat_flow - self.Heat_source) / self.Heat_nominal)



from rtctools_heat_network.pycml import Component


class HeatComponent(Component):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)


class BaseAsset(Component):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)
        self.state = 1

        self.variable_operational_cost_coefficient = 0.0
        self.fixed_operational_cost_coefficient = 0.0
        self.investment_cost_coefficient = 0.0
        self.installation_cost = 0.0



from rtctools_heat_network.pycml import Variable

from .electricity_base import ElectricityPort
from .._internal import BaseAsset
from .._internal.electricity_component import ElectricityComponent


class ElectricityDemand(ElectricityComponent, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "electricity_demand"
        self.min_voltage = 230.0

        self.add_variable(ElectricityPort, "ElectricityIn")
        self.add_variable(Variable, "Electricity_demand", min=0.0)
        self.elec_power_nominal = self.Electricity_demand.nominal

        self.add_equation((self.ElectricityIn.Power - self.Electricity_demand))



import datetime
import json
import locale
import logging
import os
import time
import xml.etree.ElementTree as ET  # noqa: N817
from pathlib import Path

from influxdb import InfluxDBClient

import numpy as np

from rtctools._internal.alias_tools import AliasDict
from rtctools.data.storage import DataStore
from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.goal_programming_mixin_base import Goal
from rtctools.optimization.linearized_order_goal_programming_mixin import (
    LinearizedOrderGoalProgrammingMixin,
)
from rtctools.optimization.single_pass_goal_programming_mixin import (
    CachingQPSol,
    SinglePassGoalProgrammingMixin,
)
from rtctools.util import run_optimization_problem

from rtctools_heat_network.esdl.esdl_mixin import ESDLMixin
from rtctools_heat_network.head_loss_mixin import HeadLossOption
from rtctools_heat_network.heat_mixin import HeatMixin
from rtctools_heat_network.workflows.goals.minimize_tco_goal import MinimizeTCO
from rtctools_heat_network.workflows.io.write_output import ScenarioOutput
from rtctools_heat_network.workflows.utils.helpers import main_decorator


DB_HOST = "172.17.0.2"
DB_PORT = 8086
DB_NAME = "Warmtenetten"
DB_USER = "admin"
DB_PASSWORD = "admin"

logger = logging.getLogger("WarmingUP-MPC")
logger.setLevel(logging.INFO)

locale.setlocale(locale.LC_ALL, "")

ns = {"fews": "http://www.wldelft.nl/fews", "pi": "http://www.wldelft.nl/fews/PI"}

WATT_TO_MEGA_WATT = 1.0e6
WATT_TO_KILO_WATT = 1.0e3


class TargetHeatGoal(Goal):
    priority = 1

    order = 2

    def __init__(self, state, target):
        self.state = state

        self.target_min = target
        self.target_max = target
        try:
            self.function_range = (-1.0e6, max(2.0 * max(target.values), 1.0e6))
            self.function_nominal = max(np.median(target.values), 1.0e6)
        except Exception:
            self.function_range = (-1.0e6, max(2.0 * target, 1.0e6))
            self.function_nominal = max(target, 1.0e6)

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(self.state)


class EndScenarioSizing(
    ScenarioOutput,
    HeatMixin,
    LinearizedOrderGoalProgrammingMixin,
    SinglePassGoalProgrammingMixin,
    ESDLMixin,
    CollocatedIntegratedOptimizationProblem,
):
    """
    Goal priorities are:
    1. minimize TCO = Capex + Opex*lifetime
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._override_hn_options = {}

        self._number_of_years = 30.0

        self.__indx_max_peak = None
        self.__day_steps = 5

        # self._override_pipe_classes = {}

        # variables for solver settings
        self._qpsol = None

        self._hot_start = False

        # Store (time taken, success, objective values, solver stats) per priority
        self._priorities_output = []
        self.__priority = None
        self.__priority_timer = None

        self.__heat_demand_bounds = dict()
        self.__heat_demand_nominal = dict()

    def _get_runinfo_path_root(self):
        runinfo_path = Path(self.esdl_run_info_path).resolve()
        tree = ET.parse(runinfo_path)
        return tree.getroot()

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)
        parameters["peak_day_index"] = self.__indx_max_peak
        parameters["time_step_days"] = self.__day_steps
        parameters["number_of_years"] = self._number_of_years
        return parameters

    def pipe_classes(self, p):
        return self._override_pipe_classes.get(p, [])

    def pre(self):
        self._qpsol = CachingQPSol()

        super().pre()

        # parameters = self.parameters(0)
        # bounds = self.bounds()

        # TODO: these constraints do no longer work as we now have varying size timesteps
        # for s in self._setpoint_constraints_sources:
        #     # Here we enforce that over the full time horizon no setpoint changes can be done
        #     self._timed_setpoints[s] = (len(self.times()), 0)
        #
        # # Mixed-interger formulation of component setpoint
        # for component_name in self._timed_setpoints.keys():
        #     # Make 1 variable per component (so not per control
        #     # variable) which represents if the setpoint of the component
        #     # is changed (1) is not changed (0) in a timestep
        #     change_setpoint_var = f"{component_name}._change_setpoint_var"
        #     self._component_to_change_setpoint_map[component_name] = change_setpoint_var
        #     self._change_setpoint_var[change_setpoint_var] = ca.MX.sym(change_setpoint_var)
        #     self._change_setpoint_bounds[change_setpoint_var] = (0, 1.0)

    def read(self):
        """
        Reads the yearly profile with hourly time steps and adapt to a daily averaged profile
        except for the day with the peak demand.
        """
        super().read()

        demands = self.heat_network_components.get("demand", [])
        new_datastore = DataStore(self)
        new_datastore.reference_datetime = self.io.datetimes[0]

        for ensemble_member in range(self.ensemble_size):
            parameters = self.parameters(ensemble_member)

            total_demand = None
            for demand in demands:
                try:
                    demand_values = self.get_timeseries(
                        f"{demand}.target_heat_demand", ensemble_member
                    ).values
                except KeyError:
                    continue
                if total_demand is None:
                    total_demand = demand_values
                else:
                    total_demand += demand_values
                max_demand = max(demand_values)
                self.__heat_demand_nominal[f"{demand}.Heat_demand"] = max_demand

            # TODO: the approach of picking one peak day was introduced for a network with a tree
            #  layout and all big sources situated at the root of the tree. It is not guaranteed
            #  that an optimal solution is reached in different network topologies.
            idx_max = int(np.argmax(total_demand))
            max_day = idx_max // 24
            nr_of_days = len(total_demand) // 24
            new_date_times = list()
            day_steps = self.__day_steps

            self.__indx_max_peak = max_day // day_steps
            if max_day % day_steps > 0:
                self.__indx_max_peak += 1.0

            for day in range(0, nr_of_days, day_steps):
                if day == max_day // day_steps * day_steps:
                    if max_day > day:
                        new_date_times.append(self.io.datetimes[day * 24])
                    new_date_times.extend(self.io.datetimes[max_day * 24 : max_day * 24 + 24])
                    if (day + day_steps - 1) > max_day:
                        new_date_times.append(self.io.datetimes[max_day * 24 + 24])
                else:
                    new_date_times.append(self.io.datetimes[day * 24])
            new_date_times.append(self.io.datetimes[-1] + datetime.timedelta(hours=1))

            new_date_times = np.asarray(new_date_times)
            parameters["times"] = [x.timestamp() for x in new_date_times]

            for demand in demands:
                var_name = f"{demand}.target_heat_demand"
                self._set_data_with_averages_and_peak_day(
                    datastore=new_datastore,
                    variable_name=var_name,
                    ensemble_member=ensemble_member,
                    new_date_times=new_date_times,
                )

            # TODO: this has not been tested but is required if a production profile is included
            #  in the data
            for source in self.heat_network_components.get("source", []):
                try:
                    self.get_timeseries(f"{source}.target_heat_source", ensemble_member)
                except KeyError:
                    logger.debug(
                        f"{source} has no production profile, skipping setting the "
                        f"production profile"
                    )
                    continue
                var_name = f"{source}.target_heat_source"
                self._set_data_with_averages_and_peak_day(
                    datastore=new_datastore,
                    variable_name=var_name,
                    ensemble_member=ensemble_member,
                    new_date_times=new_date_times,
                )

        self.io = new_datastore

        logger.info("HeatProblem read")

    def _set_data_with_averages_and_peak_day(
        self,
        datastore: DataStore,
        variable_name: str,
        ensemble_member: int,
        new_date_times: np.array,
    ):
        try:
            data = self.get_timeseries(variable=variable_name, ensemble_member=ensemble_member)
        except KeyError:
            datastore.set_timeseries(
                variable=variable_name,
                datetimes=new_date_times,
                values=np.asarray([0.0] * len(new_date_times)),
                ensemble_member=ensemble_member,
                check_duplicates=True,
            )
            return

        new_data = list()
        data_timestamps = data.times
        data_datetimes = [
            self.io.datetimes[0] + datetime.timedelta(seconds=s) for s in data_timestamps
        ]
        assert new_date_times[0] == data_datetimes[0]
        data_values = data.values

        values_for_mean = [0.0]
        for dt, val in zip(data_datetimes, data_values):
            if dt in new_date_times:
                new_data.append(np.mean(values_for_mean))
                values_for_mean = [val]
            else:
                values_for_mean.append(val)

        # last datetime is not in input data, so we need to take the mean of the last bit
        new_data.append(np.mean(values_for_mean))

        datastore.set_timeseries(
            variable=variable_name,
            datetimes=new_date_times,
            values=np.asarray(new_data),
            ensemble_member=ensemble_member,
            check_duplicates=True,
        )

    def bounds(self):
        bounds = super().bounds()
        bounds.update(self.__heat_demand_bounds)
        return bounds

    def variable_nominal(self, variable):
        try:
            return self.__heat_demand_nominal[variable]
        except KeyError:
            return super().variable_nominal(variable)

    def heat_network_options(self):
        # TODO: make empty placeholder in HeatProblem we don't know yet how to put the global
        #  constraints in the ESDL e.g. min max pressure
        options = super().heat_network_options()
        options["minimum_velocity"] = 0.0
        options["maximum_velocity"] = 3.0
        options["maximum_temperature_der"] = np.inf
        options["heat_loss_disconnected_pipe"] = False
        options["head_loss_option"] = HeadLossOption.NO_HEADLOSS
        # options.update(self._override_hn_options)
        return options

    def esdl_heat_model_options(self):
        """Overwrites the fraction of the minimum tank volume"""
        options = super().esdl_heat_model_options()
        options["min_fraction_tank_volume"] = 0.0
        return options

    def path_goals(self):
        goals = super().path_goals().copy()
        bounds = self.bounds()

        for demand in self.heat_network_components["demand"]:
            # target = self.get_timeseries(f"{demand}.target_heat_demand_peak")
            target = self.get_timeseries(f"{demand}.target_heat_demand")
            if bounds[f"{demand}.HeatIn.Heat"][1] < max(target.values):
                logger.warning(
                    f"{demand} has a flow limit lower that wat is required for the maximum demand"
                )
            state = f"{demand}.Heat_demand"

            goals.append(TargetHeatGoal(state, target))
        return goals

    def goals(self):
        goals = super().goals().copy()
        # We do a minization of TCO consisting of CAPEX and OPEX over 25 years
        # CAPEX is based upon the boolean placement variables and the optimized maximum sizes
        # Note that CAPEX for geothermal and ATES is also dependent on the amount of doublets
        # In practice this means that the CAPEX is mainly driven by the peak day problem
        # The OPEX is based on the Source strategy which is computed on the __daily_avg variables
        # The OPEX thus is based on an avg strategy and discrepancies due to fluctuations intra-day
        # are possible.
        # The idea behind the two timelines is that the optimizer can make the OPEX vs CAPEX
        # trade-offs

        goals.append(MinimizeTCO(priority=2, number_of_years=self._number_of_years))

        return goals

    def constraints(self, ensemble_member):
        constraints = super().constraints(ensemble_member)

        for a in self.heat_network_components.get("ates", []):
            stored_heat = self.state_vector(f"{a}.Stored_heat")
            constraints.append(((stored_heat[-1] - stored_heat[0]), 0.0, np.inf))

        for b in self.heat_network_components.get("buffer", {}):
            vars = self.state_vector(f"{b}.Heat_buffer")
            symbol_stored_heat = self.state_vector(f"{b}.Stored_heat")
            constraints.append((symbol_stored_heat[self.__indx_max_peak], 0.0, 0.0))
            for i in range(len(self.times())):
                if i < self.__indx_max_peak or i > (self.__indx_max_peak + 23):
                    constraints.append((vars[i], 0.0, 0.0))

        return constraints

    def history(self, ensemble_member):
        return AliasDict(self.alias_relation)

    def __state_vector_scaled(self, variable, ensemble_member):
        canonical, sign = self.alias_relation.canonical_signed(variable)
        return (
            self.state_vector(canonical, ensemble_member) * self.variable_nominal(canonical) * sign
        )

    def solver_options(self):
        options = super().solver_options()
        options["casadi_solver"] = self._qpsol
        options["solver"] = "gurobi"
        gurobi_options = options["gurobi"] = {}
        gurobi_options["MIPgap"] = 0.05
        gurobi_options["threads"] = 4

        return options

    def solver_success(self, solver_stats, log_solver_failure_as_error):
        success, log_level = super().solver_success(solver_stats, log_solver_failure_as_error)

        # Allow time-outs for CPLEX and CBC
        if (
            solver_stats["return_status"] == "time limit exceeded"
            or solver_stats["return_status"] == "stopped - on maxnodes, maxsols, maxtime"
        ):
            if self.objective_value > 1e10:
                # Quick check on the objective value. If no solution was
                # found, this is typically something like 1E50.
                return success, log_level

            return True, logging.INFO
        else:
            return success, log_level

    def priority_started(self, priority):
        self.__priority = priority
        self.__priority_timer = time.time()

        super().priority_started(priority)

    def priority_completed(self, priority):
        super().priority_completed(priority)

        self._hot_start = True

        time_taken = time.time() - self.__priority_timer
        self._priorities_output.append(
            (
                priority,
                time_taken,
                True,
                self.objective_value,
                self.solver_stats,
            )
        )

    def post(self):
        # In case the solver fails, we do not get in priority_completed(). We
        # append this last priority's statistics here in post().
        # TODO: check if we still need this small part of code below
        success, _ = self.solver_success(self.solver_stats, False)
        if not success:
            time_taken = time.time() - self.__priority_timer
            self._priorities_output.append(
                (
                    self.__priority,
                    time_taken,
                    False,
                    self.objective_value,
                    self.solver_stats,
                )
            )

        super().post()
        results = self.extract_results()
        parameters = self.parameters(0)
        bounds = self.bounds()

        for d in self.heat_network_components.get("demand", []):
            realized_demand = results[f"{d}.Heat_demand"]
            target = self.get_timeseries(f"{d}.target_heat_demand").values
            timesteps = np.diff(self.get_timeseries(f"{d}.target_heat_demand").times)
            parameters[f"{d}.target_heat_demand"] = target.tolist()
            delta_energy = np.sum((realized_demand - target)[1:] * timesteps / 1.0e9)
            if delta_energy >= 1.0:
                logger.warning(f"For demand {d} the target is not matched by {delta_energy} GJ")

        if self.esdl_run_info_path is None:
            logger.warning(
                "Not writing results since no esdl path to write the files to is specified"
            )
            return

        root = self._get_runinfo_path_root()
        parameters_dict = dict()
        workdir = root.findtext("pi:outputResultsFile", namespaces=ns)
        if workdir is None:
            logger.error("No workdir specified, skipping writing results")
            return

        parameter_path = os.path.join(workdir, "parameters.json")
        for key, value in parameters.items():
            new_value = value  # [x for x in value]
            parameters_dict[key] = new_value
        if parameter_path is None:
            workdir = root.findtext("pi:workDir", namespaces=ns)
            parameter_path = os.path.join(workdir, "parameters.json")
            if not Path(workdir).is_absolute():
                parameter_path = Path(workdir).resolve().parent
                parameter_path = os.path.join(parameter_path.__str__() + "parameters.json")
        with open(parameter_path, "w") as file:
            json.dump(parameters_dict, fp=file)

        root = self._get_runinfo_path_root()
        bounds_dict = dict()
        # bounds_path = root.findtext("pi:outputResultsFile", namespaces=ns)
        bounds_path = os.path.join(workdir, "bounds.json")
        for key, value in bounds.items():
            if "Stored_heat" not in key:
                new_value = value  # [x for x in value]
                # if len(new_value) == 1:
                #     new_value = new_value[0]
                bounds_dict[key] = new_value
        if bounds_path is None:
            workdir = root.findtext("pi:workDir", namespaces=ns)
            bounds_path = os.path.join(workdir, "bounds.json")
            if not Path(workdir).is_absolute():
                bounds_path = Path(workdir).resolve().parent
                bounds_path = os.path.join(bounds_path.__str__() + "bounds.json")
        with open(bounds_path, "w") as file:
            json.dump(bounds_dict, fp=file)

        root = self._get_runinfo_path_root()
        results_path = root.findtext("pi:outputResultsFile", namespaces=ns)
        results_dict = dict()

        for key, values in results.items():
            new_value = values.tolist()
            if len(new_value) == 1:
                new_value = new_value[0]
            results_dict[key] = new_value

        results_path = os.path.join(workdir, "results.json")
        if results_path is None:
            workdir = root.findtext("pi:workDir", namespaces=ns)
            results_path = os.path.join(workdir, "results.json")
            if not Path(workdir).is_absolute():
                results_path = Path(workdir).resolve().parent
                results_path = os.path.join(results_path.__str__() + "results.json")
        with open(results_path, "w") as file:
            json.dump(results_dict, fp=file)

        # Optimized ESDL
        self._write_updated_esdl()


def connect_database():
    client = InfluxDBClient(
        host=DB_HOST, port=DB_PORT, username=DB_USER, password=DB_PASSWORD, database=DB_NAME
    )
    if DB_NAME not in client.get_list_database():
        client.create_database(DB_NAME)
    return client


def format_datetime(dt):
    date, time = dt.split(" ")
    day, month, year = date.split("-")
    ndate = year + "-" + month + "-" + day
    ntime = time + ":00+0000"
    return ndate + "T" + ntime


class EndScenarioSizingHIGHS(EndScenarioSizing):
    def post(self):
        super().post()

        # results = self.extract_results()
        # client = connect_database()
        #
        # json_body = []
        #
        # for asset in [*self.heat_network_components.get("source", []),
        #               *self.heat_network_components.get("demand", []),
        #               *self.heat_network_components.get("pipe", []),
        #               *self.heat_network_components.get("buffer", []),
        #               *self.heat_network_components.get("ates", []),
        #               *self.heat_network_components.get("heat_exchanger", []),
        #               *self.heat_network_components.get("heat_pump", [])]:
        #     for i in range(len(self.times())):
        #         fields = {}
        #         try:
        #             # For all components dealing with one hydraulic system
        #             for variable in ["Heat_flow", "HeatIn.Q", "HeatIn.H"]:
        #                 fields[variable] = results[f"{asset}." + variable][i]
        #         except Exception:
        #             # For all components dealing with two hydraulic system
        #             for variable in ["Heat_flow", "Primary.HeatIn.Q", "Primary.HeatIn.H",
        #                              "Secondary.HeatIn.Q", "Secondary.HeatIn.H"]:
        #                 fields[variable] = results[f"{asset}." + variable][i]
        #
        #         json_body.append({
        #             "measurement": asset,
        #             "time": format_datetime(self.io.datetimes[i].strftime('%Y-%m-%d %H:%M')),
        #             "fields": fields
        #         })
        # client.write_points(points=json_body, database=DB_NAME, batch_size=100)
        self._write_updated_esdl(db_profiles=False)

    def solver_options(self):
        options = super().solver_options()
        options["casadi_solver"] = self._qpsol
        options["solver"] = "highs"
        highs_options = options["highs"] = {}
        highs_options["mip_rel_gap"] = 0.02

        options["gurobi"] = None

        return options


class EndScenarioSizingCBC(EndScenarioSizing):
    def post(self):
        super().post()

        # results = self.extract_results()
        # client = connect_database()
        #
        # json_body = []
        #
        # for asset in [*self.heat_network_components.get("source", []),
        #               *self.heat_network_components.get("demand", []),
        #               *self.heat_network_components.get("pipe", []),
        #               *self.heat_network_components.get("buffer", []),
        #               *self.heat_network_components.get("ates", []),
        #               *self.heat_network_components.get("heat_exchanger", []),
        #               *self.heat_network_components.get("heat_pump", [])]:
        #     for i in range(len(self.times())):
        #         fields = {}
        #         try:
        #             # For all components dealing with one hydraulic system
        #             for variable in ["Heat_flow", "HeatIn.Q", "HeatIn.H"]:
        #                 fields[variable] = results[f"{asset}." + variable][i]
        #         except Exception:
        #             # For all components dealing with two hydraulic system
        #             for variable in ["Heat_flow", "Primary.HeatIn.Q", "Primary.HeatIn.H",
        #                              "Secondary.HeatIn.Q", "Secondary.HeatIn.H"]:
        #                 fields[variable] = results[f"{asset}." + variable][i]
        #
        #         json_body.append({
        #             "measurement": asset,
        #             "time": format_datetime(self.io.datetimes[i].strftime('%Y-%m-%d %H:%M')),
        #             "fields": fields
        #         })
        # client.write_points(points=json_body, database=DB_NAME, batch_size=100)
        self._write_updated_esdl(db_profiles=False)

    def solver_options(self):
        options = super().solver_options()
        # options["casadi_solver"] = self._qpsol
        options["solver"] = "cbc"
        options["gurobi"] = None

        if options["solver"] == "cbc":
            options["hot_start"] = self._hot_start
            cbc_options = options["cbc"] = {}
            cbc_options["seconds"] = 300000.0

        return options


@main_decorator
def main(runinfo_path, log_level):
    logger.info("Run Scenario Sizing")
    _ = run_optimization_problem(
        EndScenarioSizingHIGHS,
        esdl_run_info_path=runinfo_path,
        log_level=log_level,
    )

    # results = solution.extract_results()


if __name__ == "__main__":
    main()



from .ates import ATES
from .buffer import Buffer
from .check_valve import CheckValve
from .control_valve import ControlValve
from .demand import Demand
from .electricity.electricity_cable import ElectricityCable
from .electricity.electricity_demand import ElectricityDemand
from .electricity.electricity_node import ElectricityNode
from .electricity.electricity_source import ElectricitySource
from .electricity.heat_pump_elec import HeatPumpElec
from .gas.gas_demand import GasDemand
from .gas.gas_node import GasNode
from .gas.gas_pipe import GasPipe
from .gas.gas_source import GasSource
from .geothermal_source import GeothermalSource
from .heat_exchanger import HeatExchanger
from .heat_four_port import HeatFourPort
from .heat_port import HeatPort
from .heat_pump import HeatPump
from .heat_two_port import HeatTwoPort
from .node import Node
from .pipe import Pipe
from .pump import Pump
from .source import Source

__all__ = [
    "ATES",
    "Buffer",
    "CheckValve",
    "ControlValve",
    "Demand",
    "ElectricityCable",
    "ElectricityDemand",
    "ElectricityNode",
    "ElectricitySource",
    "GasDemand",
    "GasNode",
    "GasPipe",
    "GasSource",
    "GeothermalSource",
    "HeatExchanger",
    "HeatFourPort",
    "HeatPort",
    "HeatPump",
    "HeatPumpElec",
    "HeatTwoPort",
    "Node",
    "Pipe",
    "Pump",
    "Source",
]



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem

from rtctools_heat_network.workflows import EndScenarioSizingHIGHS


class TestEndScenarioSizing(TestCase):
    def test_end_scenario_sizing(self):
        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a full year with timesteps of 5 days and hour timesteps
        # for the peak day
        solution = run_optimization_problem(EndScenarioSizingHIGHS, base_folder=base_folder)

        results = solution.extract_results()
        # In the future we want to check the following
        # Is the timeline correctly converted, correct peak day, correct amount of timesteps, etc.
        # Check whether expected assets are disabled
        # Check the optimal size of assets
        # Check the cost breakdown, check whether all the enabled assets are in the cost breakdown
        # Check that computation time is within expected bounds

        # Check whehter the heat demand is matched
        for d in solution.heat_network_components.get("demand", []):
            target = solution.get_timeseries(f"{d}.target_heat_demand").values
            np.testing.assert_allclose(target, results[f"{d}.Heat_demand"])

        # Check whether cyclic ates constraint is working
        for a in solution.heat_network_components.get("ates", []):
            stored_heat = results[f"{a}.Stored_heat"]
            np.testing.assert_allclose(stored_heat[0], stored_heat[-1], atol=1.0)

        # Check whether buffer tank is only active in peak day
        peak_day_indx = solution.parameters(0)["peak_day_index"]
        for b in solution.heat_network_components.get("buffer", []):
            heat_buffer = results[f"{b}.Heat_buffer"]
            for i in range(len(solution.times())):
                if i < peak_day_indx or i > (peak_day_indx + 23):
                    np.testing.assert_allclose(heat_buffer, 0.0)


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestEndScenarioSizing()
    a.test_end_scenario_sizing()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem


class TestPyCMLvsModelica(TestCase):
    def test_basic_source_and_demand_heat(self):
        import models.basic_source_and_demand.src.heat_comparison as heat_comparison
        from models.basic_source_and_demand.src.heat_comparison import HeatModelica, HeatPython

        base_folder = Path(heat_comparison.__file__).resolve().parent.parent

        case_modelica = run_optimization_problem(HeatModelica, base_folder=base_folder)
        case_python = run_optimization_problem(HeatPython, base_folder=base_folder)

        self.assertAlmostEqual(case_modelica.objective_value, case_python.objective_value, 6)

    def test_basic_source_and_demand_qth(self):
        import models.basic_source_and_demand.src.qth_comparison as qth_comparison
        from models.basic_source_and_demand.src.qth_comparison import QTHModelica, QTHPython

        base_folder = Path(qth_comparison.__file__).resolve().parent.parent

        case_modelica = run_optimization_problem(QTHModelica, base_folder=base_folder)
        case_python = run_optimization_problem(QTHPython, base_folder=base_folder)

        np.testing.assert_allclose(case_modelica._objective_values, case_python._objective_values)

    # def test_basic_buffer_example(self):
    #     from models.basic_buffer.src.compare import (
    #         HeatProblemPyCML,
    #         QTHProblemPyCML,
    #         base_folder,
    #     )
    #
    #     # TODO: this test should be moved as the comparsion with modellica is no longer relevant.
    #     #
    #     # This test should only check whether the QTH problem is seeded with a feasible set of
    #     # flow directions. Note that the current example is flawed as there seem to be multiple
    #     # equivalent MILP solutions to the problem. Which do not all give feasible directions for
    #     # QTH.
    #     pycml_heat, pycml_qth = run_heat_network_optimization(
    #         HeatProblemPyCML, QTHProblemPyCML, base_folder=base_folder
    #     )
    #     self.assertTrue(np.asarray(pycml_heat._objective_values) >= 0.0)
    #     self.assertTrue(all(np.asarray(pycml_qth._objective_values) >= 0.0))



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem


class TestAtes(TestCase):
    def test_ates(self):
        import models.test_case_small_network_with_ates.src.run_ates as run_ates
        from models.test_case_small_network_with_ates.src.run_ates import (
            HeatProblem,
        )

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a full year with 365 day timesteps
        solution = run_optimization_problem(HeatProblem, base_folder=base_folder)

        results = solution.extract_results()

        stored_heat = results["ATES_033c.Stored_heat"]
        heat_loss = results["ATES_033c.Heat_loss"]
        heat_ates = results["ATES_033c.Heat_ates"]
        summed_charge = np.sum(np.clip(heat_ates, 0.0, np.inf))
        summed_discharge = np.abs(np.sum(np.clip(heat_ates, -np.inf, 0.0)))

        # Test if the heat loss is as expected
        coeff = solution.parameters(0)["ATES_033c.heat_loss_coeff"]
        np.testing.assert_allclose(heat_loss, coeff * stored_heat)
        np.testing.assert_array_less(summed_discharge, summed_charge)

        # Test begin and end same state
        np.testing.assert_allclose(stored_heat[0], stored_heat[-1])


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestAtes()
    a.test_ates()
    temp = 0



from numpy import nan

from rtctools_heat_network.pycml import Variable
from rtctools_heat_network.pycml.component_library.heat._internal import BaseAsset
from rtctools_heat_network.pycml.component_library.heat.heat_four_port import HeatFourPort


class HeatExchanger(HeatFourPort, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(),
                modifiers,
            ),
        )

        self.component_type = "heat_exchanger"
        self.efficiency = nan

        self.nominal = (
            self.Secondary.Q_nominal * self.Secondary.rho * self.Secondary.cp * self.Secondary.dT
        )

        self.price = nan

        # Assumption: heat in/out and added is nonnegative

        self.add_variable(Variable, "Primary_heat", min=0.0)
        self.add_variable(Variable, "Secondary_heat", min=0.0)
        self.add_variable(Variable, "Heat_flow", nominal=self.nominal)
        self.add_variable(Variable, "dH_prim")
        self.add_variable(Variable, "dH_sec")

        # Hydraulically decoupled so Heads remain the same
        self.add_equation(self.dH_prim - (self.Primary.HeatOut.H - self.Primary.HeatIn.H))
        self.add_equation(self.dH_sec - (self.Secondary.HeatOut.H - self.Secondary.HeatIn.H))

        self.add_equation(
            ((self.Primary_heat * self.efficiency - self.Secondary_heat) / self.nominal)
        )

        self.add_equation(
            (
                (self.Primary_heat - (self.Primary.HeatIn.Heat - self.Primary.HeatOut.Heat))
                / self.nominal
            )
        )
        self.add_equation(
            (
                (self.Secondary_heat - (self.Secondary.HeatOut.Heat - self.Secondary.HeatIn.Heat))
                / self.nominal
            )
        )
        self.add_equation((self.Heat_flow - self.Secondary_heat) / self.nominal)



import logging
import math
from typing import Dict, Tuple, Type

import esdl

from rtctools_heat_network.pycml import SymbolicParameter
from rtctools_heat_network.pycml.component_library.qth import (
    Buffer,
    CheckValve,
    ControlValve,
    Demand,
    GeothermalSource,
    Node,
    Pipe,
    Pump,
    Source,
)

from .asset_to_component_base import MODIFIERS, _AssetToComponentBase
from .common import Asset
from .esdl_model_base import _ESDLModelBase

logger = logging.getLogger("rtctools_heat_network")


class AssetToQTHComponent(_AssetToComponentBase):
    def __init__(
        self,
        theta,
        *args,
        v_nominal=1.0,
        v_max=5.0,
        minimum_temperature=10.0,
        maximum_temperature=110.0,
        rho=988.0,
        cp=4200.0,
        min_fraction_tank_volume=0.05,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.theta = theta
        self.v_nominal = v_nominal
        self.v_max = v_max
        self.minimum_temperature = minimum_temperature
        self.maximum_temperature = maximum_temperature
        self.rho = rho
        self.cp = cp
        self.min_fraction_tank_volume = min_fraction_tank_volume

    @property
    def _rho_cp_modifiers(self):
        return dict(rho=self.rho, cp=self.cp)

    def convert_buffer(self, asset: Asset) -> Tuple[Type[Buffer], MODIFIERS]:
        assert asset.asset_type == "HeatStorage"

        (
            supply_temperature,
            return_temperature,
            supply_temperature_id,
            return_temperature_id,
        ) = self._get_supply_return_temperatures(asset)

        heat_to_discharge_fac = 1 / (self.rho * self.cp * (supply_temperature - return_temperature))

        # Assume that:
        # - the capacity is the relative heat that can be stored in the buffer;
        # - the tanks are always at least `min_fraction_tank_volume` full;
        # - same height as radius to compute dimensions.
        min_fraction_tank_volume = self.min_fraction_tank_volume

        if asset.attributes["capacity"] and asset.attributes["volume"]:
            logger.warning(
                f"{asset.asset_type} '{asset.name}' has both capacity and volume specified. "
                f"Volume with value of {asset.attributes['volume']} m3 will be used."
            )

        capacity = 0.0
        if asset.attributes["volume"]:
            capacity = (
                asset.attributes["volume"]
                * self.rho
                * self.cp
                * (supply_temperature - return_temperature)
            )
        elif asset.attributes["capacity"]:
            capacity = asset.attributes["capacity"]
        else:
            logger.error(
                f"{asset.asset_type} '{asset.name}' has both not capacity and volume specified. "
                f"Please specify one of the two"
            )

        r = (capacity * (1 + min_fraction_tank_volume) * heat_to_discharge_fac / math.pi) ** (
            1.0 / 3.0
        )

        # Note that these flow constraints are estimations based on the
        # carrier temperatures, same way as CHESS handles them.
        hfr_charge_max = asset.attributes.get("maxChargeRate", math.inf) or math.inf
        hfr_discharge_max = asset.attributes.get("maxDischargeRate", math.inf) or math.inf
        q_charge_max = hfr_charge_max * heat_to_discharge_fac
        q_discharge_max = hfr_discharge_max * heat_to_discharge_fac

        modifiers = dict(
            Q_nominal=self._get_connected_q_nominal(asset),
            height=r,
            radius=r,
            heat_transfer_coeff=1.0,
            min_fraction_tank_volume=min_fraction_tank_volume,
            init_T_hot_tank=supply_temperature,
            init_T_cold_tank=return_temperature,
            Q_hot_pipe=dict(min=-q_discharge_max, max=q_charge_max),
            Q_cold_pipe=dict(min=-q_discharge_max, max=q_charge_max),
            **self._supply_return_temperature_modifiers(asset),
            **self._rho_cp_modifiers,
        )

        return Buffer, modifiers

    def convert_demand(self, asset: Asset) -> Tuple[Type[Demand], MODIFIERS]:
        assert asset.asset_type in {"GenericConsumer", "HeatingDemand"}

        # TODO: Why is the default zero for both, that's just weird. What if I
        # actually want a minimum of 0.0 (instead of treating it as a
        # NaN/None/not specified)
        # TODO: Are these even the correct values to use? It does not
        # say anything about whether this is the min/max on the feed or return line?
        # currently assuming they are the min/max of the _feed_ line.

        minimum_temperature = asset.attributes["minTemperature"]
        if minimum_temperature == 0.0:
            logger.warning(
                f"{asset.asset_type} '{asset.name}' has an unspecified minimum temperature. "
                f"Using default value of {self.minimum_temperature}."
            )
            minimum_temperature = self.minimum_temperature

        maximum_temperature = asset.attributes["maxTemperature"]
        if maximum_temperature == 0.0:
            logger.warning(
                f"{asset.asset_type} '{asset.name}' has an unspecified maximum temperature. "
                f"Using default value of {self.maximum_temperature}."
            )
            maximum_temperature = self.maximum_temperature

        (
            supply_temperature,
            return_temperature,
            supply_temperature_id,
            return_temperature_id,
        ) = self._get_supply_return_temperatures(asset)
        max_demand = asset.attributes["power"] if asset.attributes["power"] else math.inf

        modifiers = dict(
            theta=self.theta,
            Q_nominal=self._get_connected_q_nominal(asset),
            QTHIn=dict(T=dict(min=minimum_temperature, max=maximum_temperature)),
            Heat_demand=dict(min=0.0, max=max_demand),
            **self._supply_return_temperature_modifiers(asset),
            **self._rho_cp_modifiers,
        )

        return Demand, modifiers

    def convert_node(self, asset: Asset) -> Tuple[Type[Node], MODIFIERS]:
        assert asset.asset_type == "Joint"  # TODO: temperature?

        sum_in = 0
        sum_out = 0

        for x in asset.attributes["port"].items:
            if isinstance(x, esdl.esdl.InPort):
                sum_in += len(x.connectedTo)
            if isinstance(x, esdl.esdl.OutPort):
                sum_out += len(x.connectedTo)

        # TODO: what do we want if no carrier is specified.
        carrier = asset.global_properties["carriers"][asset.in_ports[0].carrier.id]
        if carrier["__rtc_type"] == "supply":
            temp = carrier["supplyTemperature"]
        elif carrier["__rtc_type"] == "return":
            temp = carrier["returnTemperature"]
        else:
            temp = 50.0
        modifiers = dict(
            n=sum_in + sum_out,
            temperature=temp,
        )

        return Node, modifiers

    def convert_pipe(self, asset: Asset) -> Tuple[Type[Pipe], MODIFIERS]:
        assert asset.asset_type == "Pipe"

        (
            supply_temperature,
            return_temperature,
            supply_temperature_id,
            return_temperature_id,
        ) = self._get_supply_return_temperatures(asset)

        # NaN means the default values will be used
        insulation_thicknesses = math.nan
        conductivies_insulation = math.nan

        if "_ret" in asset.attributes["name"]:
            temperature = return_temperature
        else:
            temperature = supply_temperature

        (
            diameter,
            insulation_thicknesses,
            conductivies_insulation,
        ) = self._pipe_get_diameter_and_insulation(asset)

        area = math.pi * diameter**2 / 4.0
        q_nominal = self.v_nominal * area
        q_max = self.v_max * area

        self._set_q_nominal(asset, q_nominal)

        # TODO: We can do better with the temperature bounds.
        # Maybe global ones (temperature_supply_max / min, and temperature_return_max / min?)
        modifiers = dict(
            length=asset.attributes["length"],
            diameter=diameter,
            temperature=temperature,
            disconnectable=self._is_disconnectable_pipe(asset),
            Q=dict(min=-q_max, max=q_max, nominal=q_nominal),
            QTHIn=dict(T=dict(min=self.minimum_temperature, max=self.maximum_temperature)),
            QTHOut=dict(T=dict(min=self.minimum_temperature, max=self.maximum_temperature)),
            insulation_thickness=insulation_thicknesses,
            conductivity_insulation=conductivies_insulation,
            **self._supply_return_temperature_modifiers(asset),
            **self._rho_cp_modifiers,
        )

        return Pipe, modifiers

    def convert_pump(self, asset: Asset) -> Tuple[Type[Pump], MODIFIERS]:
        assert asset.asset_type == "Pump"

        # TODO: Maximum pump head should come from ESDL component
        # specification. Alpha-2 CHESS release only checks for capacity, head
        # is always assumed to be realizable.
        modifiers = dict(
            Q=dict(
                min=0.0,
                max=asset.attributes["pumpCapacity"] / 3600.0,
                nominal=asset.attributes["pumpCapacity"] / 7200.0,
            ),
            dH=dict(min=0.0),
        )

        return Pump, modifiers

    def convert_source(self, asset: Asset) -> Tuple[Type[Source], MODIFIERS]:
        assert asset.asset_type in {
            "GasHeater",
            "GenericProducer",
            "GeothermalSource",
            "ResidualHeatSource",
        }

        (
            supply_temperature,
            return_temperature,
            supply_temperature_id,
            return_temperature_id,
        ) = self._get_supply_return_temperatures(asset)

        # TODO: Why is the default zero for both, that's just weird. What if I
        # actually want a minimum of 0.0 (instead of treating it as a
        # NaN/None/not specified)
        # TODO: Are these even the correct values to use? It does not
        # say anything about whether this is the min/max on the feed or return line?
        # currently assuming they are the min/max of the _feed_ line.
        if asset.asset_type == "GenericProducer":
            minimum_temperature = supply_temperature
            maximum_temperature = supply_temperature
        else:
            minimum_temperature = asset.attributes["minTemperature"]
            if minimum_temperature == 0.0:
                logger.warning(
                    f"{asset.asset_type} '{asset.name}' has an unspecified minimum temperature. "
                    f"Using default value of {self.minimum_temperature}."
                )
                minimum_temperature = self.minimum_temperature

            maximum_temperature = asset.attributes["maxTemperature"]
            if maximum_temperature == 0.0:
                logger.warning(
                    f"{asset.asset_type} '{asset.name}' has an unspecified maximum temperature. "
                    f"Using default value of {self.maximum_temperature}."
                )
                maximum_temperature = self.maximum_temperature

        # get price per unit of energy,
        # assume cost of 1. if nothing is given (effectively heat loss minimization)
        price = 1.0
        if "costInformation" in asset.attributes.keys():
            if hasattr(asset.attributes["costInformation"], "variableOperationalCosts"):
                if hasattr(asset.attributes["costInformation"].variableOperationalCosts, "value"):
                    price = asset.attributes["costInformation"].variableOperationalCosts.value

        max_supply = asset.attributes["power"]
        if not max_supply:
            logger.error(f"{asset.asset_type} '{asset.name}' has no max power specified. ")
        assert max_supply > 0.0

        modifiers = dict(
            theta=self.theta,
            Q_nominal=self._get_connected_q_nominal(asset),
            QTHOut=dict(T=dict(min=minimum_temperature, max=maximum_temperature)),
            price=price,
            Heat_source=dict(min=0.0, max=max_supply, nominal=max_supply / 2.0),
            **self._supply_return_temperature_modifiers(asset),
            **self._rho_cp_modifiers,
        )

        if asset.asset_type == "GeothermalSource":
            # Note that the ESDL target flow rate is in kg/s, but we want m3/s
            try:
                modifiers["target_flow_rate"] = asset.attributes["flowRate"] / self.rho
            except KeyError:
                logger.warning(
                    f"{asset.asset_type} '{asset.name}' has no desired flow rate specified. "
                    f"'{asset.name}' will not be actuated in a constant manner"
                )
            return GeothermalSource, modifiers
        else:
            return Source, modifiers

    def convert_control_valve(self, asset: Asset) -> Tuple[Type[ControlValve], MODIFIERS]:
        assert asset.asset_type == "Valve"

        return ControlValve, {}

    def convert_check_valve(self, asset: Asset) -> Tuple[Type[CheckValve], MODIFIERS]:
        assert asset.asset_type == "CheckValve"

        return CheckValve, {}


class ESDLQTHModel(_ESDLModelBase):
    _converter_class: _AssetToComponentBase = None

    def __init__(self, assets: Dict[str, Asset], converter_class=AssetToQTHComponent, **kwargs):
        super().__init__(None)

        self.add_variable(SymbolicParameter, "theta")

        converter = converter_class(theta=self.theta, **kwargs)

        self._esdl_convert(converter, assets, "QTH")



from numpy import nan

from rtctools_heat_network.pycml import Variable

from ._internal.heat_component import BaseAsset
from .heat_two_port import HeatTwoPort


class _NonStorageComponent(HeatTwoPort, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.Q_nominal = 1.0
        self.T_supply = nan
        self.T_return = nan
        self.T_supply_id = -1
        self.T_return_id = -1
        self.dT = self.T_supply - self.T_return
        self.cp = 4200.0
        self.rho = 988.0

        # NOTE: We move a factor of 100.0 of the heat to the state entry, to
        # reduce the coefficient in front of the heat variables. This
        # particularly helps the scaling/range of the constraints that relate
        # the heat loss (if it is variable/optional) to the heat in- and out
        # of a component.
        self.Heat_nominal = self.cp * self.rho * self.dT * self.Q_nominal / 100.0

        self.add_variable(Variable, "Heat_in", nominal=self.Heat_nominal)
        self.add_variable(Variable, "Heat_out", nominal=self.Heat_nominal)

        self.add_variable(Variable, "Q", nominal=self.Q_nominal)

        self.add_variable(Variable, "H_in")
        self.add_variable(Variable, "H_out")

        self.add_variable(Variable, "Heat_flow", nominal=self.Heat_nominal)

        self.add_equation((self.Heat_out - self.HeatOut.Heat) / self.Heat_nominal)
        self.add_equation((self.Heat_in - self.HeatIn.Heat) / self.Heat_nominal)

        self.add_equation(self.HeatIn.Q - self.Q)
        self.add_equation(self.HeatIn.Q - self.HeatOut.Q)

        self.add_equation(self.HeatIn.H - self.H_in)
        self.add_equation(self.HeatOut.H - self.H_out)
